/******************************************************************************
    DTMF RECEIVER  (SIMULA UM CIRCUITO INTEGRADO MT8870DE)

    C/ microcontrolador PIC16F676

    Receptor para sinais DTMF com 4 sada codificadas em hexadecimal.

    Autor: Cludio Lrios

    Baseado nos seguintes artigos publicados em Websites originalmente escritos
    em linguagem assembly  e sendo adaptado para linguagem C:

    Adaptaes de Tabelas e algoritimos Goertzel: Radu Constantinescu.
    Microchip Note Applications AN257 'DTMF Detection Using PIC18
    Microcontrollers' by Gaurang Kavaiya.
    DTMF Decoding Software with a PIC - by Wim de Vree PE1GRL - pe1grl@amsat.org
    Fontes adicionais de pesquisas:
    Introduction: DTMF Touch Tone Decoder Using Microchip PIC Microprocessor by
    LargeMouthBass - https://www.instructables.com/member/LargeMouthBass/
    Black DTMF Decoding Algorithm - www.RomanBlack.com
    Agradecimentos a todos os autores por compartilhar seus artigos a comunidade.


    Compilador: CCS C V3.200


    PINAGEM DO PIC 16F676:
    PINO 6   =   (sada)STD (nivel '1' ao receber tons DTMF)
    PINO 12  =   (entrada) sinal DTMF
    PINO 10  =   Q1 (sada)
    PINO  9  =   Q2 (sada)
    PINO  8  =   Q3 (sada)
    PINO  7  =   Q4 (sada)
    PINO  2  =   xtal 20MHZ
    PINO  3  =   xtal 20MHZ
    PINO  1  =   VCC
    PINO 14  =   GND

    Uso didtico

    Blog Picsource
******************************************************************************/

#include <16F676.h>
#use delay(clock=20000000)//frequencia do xtal
#fuses NOWDT,HS, NOCPD, NOPROTECT, NOMCLR, NOPUT, NOBROWNOUT

//bytes

#byte   FSR         = 0x04
#byte   PCLATH      = 0x0A
#byte   STATUS      = 0X03
#byte   TRISA       = 0x85
#byte   PORTA       = 0x05
#byte   TRISC       = 0x87
#byte   PORTC       = 0x07
#byte   TMR1H       = 0x0F
#byte   TMR1L       = 0x0E
#byte   TMR0        = 0X01
#byte   OPTION_REG  = 0x81
#byte   WPU         = 0x95
#byte   CMCON       = 0x19
#byte   ANSEL       = 0x9F
#byte   ADRESH      = 0x1E
#byte   ADRESL      = 0x9E
#byte   T1CON       = 0X10
#byte   IOC         = 0x96
#byte   INTCON      = 0X0B
#byte   VRCON       = 0X99
#byte   ADCON0      = 0X1F
#byte   VRCON       = 0X99
#byte   FSR         = 0X04
#byte   INDF        = 0X00

//bit
#bit    GIE         = 0xb.7
#bit    PEIE        = 0xb.6
#bit    T0IE        = 0xb.5
#bit    TMR1IE      = 0x8C.0
#bit    TMR1IF      = 0xC.0
#bit    TMR1_ON     = 0x10.0
#bit    T0IF        = 0XB.2
#bit    TMR0_F0     = 0x01.0
#bit    C_OUT       = 0x19.6
#bit    GO_DONE     = 0X1F.1

//pinos
#bit std  = 0x7.4  //pino 6 (std=1 somente quando recebe um tom dtmf )
#bit q1   = 0x7.0  //pino 10 (segue o mesmo padro de sada do MT8870DE)
#bit q2   = 0x7.1  //pino 9
#bit q3   = 0x7.2  //pino 8
#bit q4   = 0x7.3  //pino 7


#define N_AMOSTRAGENS 240
#define N_AMOSTRAGENS_X2 113 //1/2 amostragem - 7
#define LIMITE_MIN_X1L  60  //1/4 amostragem x1
#define LIMITE_MIN_X1H  48  //1/5 amostragem x1
#define LIMITE_MIN_X2L  37  //1/3 amostragem x2 - 3
#define LIMITE_MIN_X2H  37  //1/3 amostragem x2 - 3



// variveis globais
char cnt_amostragem=0;
char mascara;
char cnt_desl;
char tecla_hex;
char tecla_hex_ant=0;
unsigned int1 fvez=0;

/* Acumulador de 32 bytes para guardar valores dos senos e cossenos
das frequncia DTMF -> 697,770,852,941 (baixas) e 1209,1336,1447,1633 (altas)
obtidas de 2 tabelas na flash: tab_low_band e tab_high_band.
Os primeiros 16 bytes contero valores de senos e cossenos para freq x 1.
Os 16 bytes seguintes contero valores de senos e cossenos para freq x 2.
Acesso ao array por nmero 'par' obteremos os valores de senos e 'impares' os
valores de cossenos. Veja abaixo:

sin_cos[0]; -> seno 697 freq x1
sin_cos[1]; -> cosseno 697 freq x1
sin_cos[2]; -> seno 770 freq x1
sin_cos[3]; -> cosseno 770 freq x1
sin_cos[4]; -> seno 852 freq x1
sin_cos[5]; -> cosseno 852 freq x1
sin_cos[6]; -> seno 941 freq x1
sin_cos[7]; -> cosseno 941 freq x1

sin_cos[8]; -> seno 1209 freq x1
sin_cos[9]; -> cosseno 1209 freq x1
sin_cos[10]; -> seno 1336 freq x1
sin_cos[11]; -> cosseno 1336 freq x1
sin_cos[12]; -> seno 1447 freq x1
sin_cos[13]; -> cosseno 1447 freq x1
sin_cos[14]; -> seno 1633 freq x1
sin_cos[15]; -> cosseno 1633 freq x1

sin_cos[16]; -> seno 697 freq x2
sin_cos[17]; -> cosseno 697 freq x2
sin_cos[18]; -> seno 770 freq x2
sin_cos[19]; -> cosseno 770 freq x2
sin_cos[20]; -> seno 852 freq x2
sin_cos[21]; -> cosseno 852 freq x2
sin_cos[22]; -> seno 941 freq x2
sin_cos[23]; -> cosseno 941 freq x2

sin_cos[24]; -> seno 1209 freq x2
sin_cos[25]; -> cosseno 1209 freq x2
sin_cos[26]; -> seno 1336 freq x2
sin_cos[27]; -> cosseno 1336 freq x2
sin_cos[28]; -> seno 1447 freq x2
sin_cos[29]; -> cosseno 1447 freq x2
sin_cos[30]; -> seno 1633 freq x2
sin_cos[31]; -> cosseno 1633 freq x2
*/
unsigned int8 sin_cos[32];//array de senos e cossenos


/* tabela de mscaras para as amostragens 697,770,852,941 hz
   Obs. Cada bit tem informao para cada uma das 4 frequncias DTMF inferiores,
   sendo:
   b7= sin 697, b6=cos697, b5=sin770, b4=cos770,
   b3=sin852, b2=cos852, b1=sin941, b0=cos941 */
const char tab_low_band[]={
 0x55,0xff,0xff,0xfa,0xaa,0xaa,0xa0,0x80,0x01,0x05,
 0x15,0x57,0x5f,0x7e,0x7e,0xfa,0xe8,0xe8,0xa0,0xa1,
 0x81,0x87,0x07,0x1f,0x1e,0x5e,0x78,0x78,0x78,0xe1,
 0xe1,0xe1,0x87,0x87,0x86,0x9e,0x1e,0x1c,0x18,0x79,
 0x79,0x71,0x63,0xe7,0xe7,0xc6,0x8e,0x8c,0x9c,0x98,
 0x19,0x39,0x33,0x73,0x63,0x66,0x66,0xe6,0xcc,0xcc,
 0x8d,0x99,0x99,0x9b,0x33,0x32,0x36,0x66,0x64,0x6c,
 0x4c,0xcd,0xc9,0xdb,0x9b,0x93,0x92,0xb2,0x36,0x36,
 0x26,0x6d,0x6d,0x4d,0x4b,0xcb,0xd2,0xd2,0x92,0xb4,
 0xb4,0xb5,0x2d,0x2d,0x2f,0x6b,0x4a,0x4a,0x42,0xd0,
 0xd0,0xd4,0xb5,0xb5,0xbf,0xaf,0x2f,0x2a,0x0a,0x40,
 0x40,0x50,0x55,0xd5,0xd5,0xff,0xbf,0xbe,0xaa,0xaa,
 0x28,0x00,0x01,0x41,0x55,0x57,0x57,0xff,0xfe,0xfa,
 0xf8,0xa8,0xa0,0xa1,0x01,0x07,0x07,0x57,0x5e,0x5e,
 0x7e,0xf8,0xf8,0xe9,0xa1,0xa1,0x83,0x87,0x06,0x06,
 0x1e,0x5c,0x58,0x78,0x79,0xf1,0xe3,0xe3,0xa7,0x96,
 0x86,0x8c,0x1c,0x1c,0x19,0x79,0x79,0x73,0x73,0xe2,
 0xe6,0xe6,0x8c,0x8c,0x8d,0x99,0x19,0x1b,0x33,0x73,
 0x72,0x66,0x64,0xe4,0xcc,0xcd,0x8d,0x8b,0x9b,0x9b,
 0x12,0x32,0x34,0x74,0x64,0x6d,0x6d,0xcd,0xcb,0xcb,
 0x9a,0x92,0x92,0xb0,0x34,0x35,0x35,0x6d,0x6f,0x6f,
 0x4b,0xca,0xca,0xd0,0x90,0x94,0xb5,0xb5,0x3f,0x2f,
 0x2f,0x6a,0x4a,0x4a,0x40,0xc0,0xd1,0xd5,0x95,0xb7,
 0xbf,0xbe,0x2e,0x2a,0x28,0x40,0x40,0x41,0x55,0xd7,
 0xd7,0xff,0xbe,0xbe,0xb8,0xa8,0x28,0x21,0x01,0x41,
 0x47,0x57,0x56,0xde,0xfe
};

/* tabela de mscaras para as amostragens 1209,1336,1447,1633 hz
   Obs. Cada bit tem informao para cada uma das 4 frequncias DTMF superiores,
   sendo:
   b7= sin 1209, b6=cos1209, b5=sin1336, b4=cos1336,
   b3=sin1477, b2=cos1477, b1=sin1633, b0=cos1633 */
const int8 tab_high_band[]={
 0x55,0xff,0xea,0xa8,0x80,0x05,0x17,0x5f,0x7e,0xf8,
 0xe0,0x81,0x87,0x17,0x1e,0x78,0x78,0xe1,0xe3,0x87,
 0x8e,0x1c,0x38,0x79,0x63,0xe7,0xc6,0x8c,0x9c,0x19,
 0x33,0x73,0x66,0xc4,0xcc,0x99,0x9b,0x33,0x32,0x64,
 0x6c,0xcd,0xcb,0x93,0xb2,0x34,0x24,0x6d,0x4b,0xcb,
 0xd2,0x90,0xb4,0x2d,0x2f,0x4b,0x4a,0xd0,0xd4,0xb5,
 0xbf,0x2f,0x2a,0x40,0x50,0xd5,0xf7,0xbf,0xaa,0x28,
 0x00,0x41,0x57,0xdf,0xfe,0xa8,0xa8,0x01,0x07,0x57,
 0x5e,0xf8,0xf8,0xa1,0xa3,0x07,0x1e,0x5c,0x78,0xf9,
 0xe3,0xa7,0x86,0x0c,0x1c,0x59,0x73,0xe3,0xe6,0x84,
 0x8c,0x19,0x1b,0x73,0x72,0x64,0xcc,0xcd,0x9b,0x9b,
 0x32,0x34,0x64,0x6d,0xcf,0xcb,0x92,0xb0,0x34,0x25,
 0x6f,0x4b,0xca,0xd0,0x94,0xb5,0x3f,0x2f,0x4a,0x40,
 0xd0,0xd5,0xb7,0xbf,0x2a,0x28,0x40,0x41,0xd7,0xff,
 0xbe,0xa8,0x28,0x01,0x47,0x57,0xde,0xfc,0xb8,0xa1,
 0x03,0x06,0x56,0x5c,0xf9,0xf9,0xa3,0xa2,0x06,0x1c,
 0x5d,0x79,0xf3,0xe2,0xa6,0x84,0x0d,0x19,0x5b,0x72,
 0xe2,0xe4,0x8d,0x8d,0x1b,0x1a,0x72,0x74,0xe5,0xed,
 0x8f,0x9a,0x12,0x30,0x75,0x65,0xef,0xca,0x8a,0x90,
 0x11,0x35,0x6f,0x6e,0xca,0xc8,0x91,0x95,0x37,0x3e,
 0x6e,0x68,0xc1,0xd1,0x97,0xbe,0x3e,0x28,0x69,0x41,
 0x47,0xd6,0xfe,0xbc,0xa9,0x21,0x03,0x46,0x56,0xdc,
 0xf9,0xb9,0xa3,0x02,0x06,0x5c,0x5d,0xf9,0xfb,0xa2,
 0xa6,0x04,0x0d,0x5d,0x7b,0xf2,0xe2,0xa4,0x85,0x0d,
 0x1b,0x5a,0x72,0xf4,0xe5,0x8d,0x8f,0x1a,0x12,0x70,
 0x75,0xe5,0xef,0x8a,0x8a
};


/* valores de magnitudes referentes a cada par de frequncias de tons DTMF
   sendo :
   bit7=1633  bit6=1447 bit5=1336 bit4=1209
   bit3=941   bit2=852  bit1=770  bit0=697
   Obs: O bit setado indica presena da respectiva frequncia conforme acima */
const char tab_magnitudes[]={
 //0   1    2    3     4   5    6    7    8    9    *   #     A    B    C    D
  40, 17,  33,  65,   18,  34,  66, 20,   36,  68,  24, 72,  129, 130,  132, 136
};

/* Tabela p/ retorno de valores p/ sadas q1...q4 (retorna conforme posies da
tabela de magnitudes */
const char tab_hex[]={ //retorna valor em hexadecimal
//                                                 *      #
//'0','1','2','3', '4', '5',  '6', '7', '8', '9', 0x2a, 0x23,'A','B', 'C', 'D'
0x0a,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0b,0x0c,0x0d,0x0e,0x0f,0x00
};

/*==============================================================================
 ROTINA DE INTERRUPO DO TIMER 0 - (executada a cada 102,4 us)
 Obs. O modo de interrupo global no tem salvamento automtico de valores dos
 principais registradores e nem apaga flag de overflow do timer0. Porque usa-la?
 O motivo  que a rotina '#int_timer0' gasta muito tempo salvando registradores
 desnecessrios aos objetivos desta montagem.
 Para funcionar, a rotina de interrupo do timer0 no pode exceder 102,4 us.
 A rotina atual gasta 96,8 us.
 Note que as rotinas usam o FSR diretamente como ponteiro. Tambm foi priorizado
 o uso de rotinas curtas. Cada rotina foi reduzida ao mximo para no exceder
 o tempo mximo de execuo.
 Se fosse usado '#int_timer0' e o modo padro de rotinas C, este tempo passaria
 tranquilamente dos 160 us.
 Cuidado especial deve ser dado a este fato caso venha a migrar para outros
 tipos de microcontroladores.
 Tambm tem que ser usado a frequncia de 20mhz (frequncia mximo deste PIC).
==============================================================================*/
#INT_GLOBAL
void isr_timer0(){
//variveis locais
char cnt_amostragem2x;
char modelo_freq;//guarda modelo da amostragem para uma dada frequencia
char STATUS_TEMP,PCLATH_SAVE,FSR_SAVE,W_TEMP;

   #ASM
   movwf	 W_TEMP
	swapf	 STATUS,W
	clrf	 STATUS		//ENSURE BANK0 SAVE
	movwf	 STATUS_TEMP
	movf	 PCLATH,W	  //SAVE PCLATH
	movwf  PCLATH_SAVE
	movf	 FSR, W
	movwf	 FSR_SAVE
   #ENDASM

  /*testa bit de sada do comparador analgico que contm a informao do sinal
  DTMF (na entrada do pino 6).
  */
  mascara=0;//inicia zerado
  if(C_OUT)mascara--;//Se sada do comparador analgico  '1' ento mascara=0xff
  cnt_amostragem2x=cnt_amostragem<<1;//obtem posio para tabela x2

  /* Obtem valores diferentes de senos e cossenos das frequncias de tom para cada
  amostragem efetuada atravs de 2 tabelas na flash. Estes valores so mascarados
  conforme a situao da entrada de sinal DTMF (pino 12, via comparador analgico).
  */


  FSR= sin_cos; //aponta ponteiro para inicio do acumulador seno/cosseno 32
  modelo_freq=tab_low_band[cnt_amostragem] ^ mascara;//obtem valores das tabela low x1
  do{
    if(modelo_freq & 0x80)INDF--;//bit =1 ento decrementa acumulador apontado
    if(!(modelo_freq & 0x80))INDF++;//bit 0 ento incrementa acumulador apontado
    modelo_freq= modelo_freq <<1; //desloca para esquerda para mudar o bit
    FSR++;//muda o acumulador
    }while(FSR<sin_cos+8);

  modelo_freq=tab_high_band[cnt_amostragem] ^ mascara;//obtem valores das tabela high x1
  do{
    if(modelo_freq & 0x80)INDF--; //se 1 decrementa
    if(!(modelo_freq & 0x80))INDF++;
    modelo_freq= modelo_freq <<1; //desloca para esquerda
    FSR++;
  }while(FSR<sin_cos+16);


  if(cnt_amostragem<N_AMOSTRAGENS_X2){//limita quantidade de amostragens em frequncias 2x
     modelo_freq=tab_low_band[cnt_amostragem2x] ^ mascara;//obtem valores das tabela low x2
     do{
      if(modelo_freq & 0x80)INDF--; //se 1 decrementa
      if(!(modelo_freq & 0x80))INDF++; //se 0 incrementa
      modelo_freq= modelo_freq <<1; //desloca para esquerda
      FSR++;
     }while(FSR<sin_cos+24);
     modelo_freq=tab_high_band[cnt_amostragem2x] ^ mascara;//obtem valores das tabela high x2
     do{
      if(modelo_freq & 0x80)INDF--; //se 1 decrementa
      if(!(modelo_freq & 0x80))INDF++; //se 0 incrementa
      modelo_freq= modelo_freq <<1; //desloca para esquerda
      FSR++;
     }while(FSR<sin_cos+32);
   }


  if(++cnt_amostragem>=N_AMOSTRAGENS){//incrementa e se chegou no valor mximo
    T0IE=0;//Desliga interrupo do timer 0
  }

  T0IF=0;//reseta flag de overflow do timer0

  #asm //restabelece principais registradores com valores anteriores
  	 movf	   FSR_SAVE,W
	 movwf	FSR
	 movf	   PCLATH_SAVE,W
	 movwf	PCLATH		// RESTORE PCLATH
	 swapf	STATUS_TEMP,W
	 movwf	STATUS
	 swapf	W_TEMP,F
	 swapf	W_TEMP,W
  #endasm
}//fim da rotina de interrupo do timer 0


/*------------------------------------------------------------------------------
     ROTINA PARA ZERAR ARRAY COM VALORES DE SENOS E COSSENOS
-------------------------------------------------------------------------------*/
void zera_array(){
  for(FSR=sin_cos;FSR<=sin_cos+32;FSR++){
     INDF=0; //apaga array seno-cosseno
  }
}

/*------------------------------------------------------------------------------
   ROTINA PARA OBTER OS TONS DTMF (SE PRESENTES NA ENTRADA DE SINAL - PINO 12)
-------------------------------------------------------------------------------*/

char decodifica_dtmf(){
char a,b,c,d;
char res_magx1,res_magx2;

   res_magx1=0;//registrador p/ resultados das magnitudes das 8 frequencias x 1
   res_magx2=0;//registrador p/ resultados das magnitudes das 8 frequencias x 2

   for(b=0;b<32;b++){

     d= sin_cos[b];
     if(d & 0x80)d = 0 - d; //obtem o valor absoluto do seno
     b++;
     c= sin_cos[b];
     if(c & 0x80)c = 0 - c; //obtem o valor absoluto do cosseno
     c= c + d;// soma seno e cosseno absolutos

     /*=========================================================================
     Rotina abaixo compara as magnitudes obtidas nos acumuladores com os
      limites mnimos e passando os valores aprovados p/ os bytes res_magx1 e
      res_magx2
    ========================================================================= */
     if(b<8){//quando b= 1,3,5,7  (low freq x1)
       res_magx1=res_magx1>>1;
       if(c>=LIMITE_MIN_X1L)res_magx1=res_magx1 | 0x80;//se c>limite
     }
     else if(b<16){//quando b= 9,11,13,15  (high freq x1)
       res_magx1=res_magx1>>1;
       if(c>=LIMITE_MIN_X1H)res_magx1=res_magx1 | 0x80;//se c>limite
     }

     else if(b<24){//quando b= 17,19,21,23 (low freq x2)
       res_magx2=res_magx2>>1;
       if(c<LIMITE_MIN_X2L)res_magx2=res_magx2 | 0x80;//se c<=limite
     }
     else {//quando b = 25,27,29,31 (high freq x2)
       res_magx2=res_magx2>>1;
       if(c<LIMITE_MIN_X2H)res_magx2=res_magx2 | 0x80;//se c<=limite
     }
   }

   res_magx1= res_magx1 & res_magx2;//faz um and lgico entre resultados x1 e x2

   /* res_magx1 contm o valor das 2 frequncias dtmf do sinal de entrada com as
   maiores magnitudes. A rotina abaixo pesquisa em uma tabela o valor da tecla
   apertada e aciona as respectivas sada q1 ... q4   */
   for(a=0;a<16;a++){//
     if(res_magx1 == tab_magnitudes[a]){//compara com tabela
       tecla_hex=tab_hex[a];//recebe valor em hexadecimal referente a tecla acionada
       if(!fvez){
         fvez=1;
         tecla_hex_ant=tecla_hex;//recebe valor 1vez
         return 0;//tom no encontrado        
       }
       else{
         fvez=0;//ajusta  saidas q1...q4 conforme valor encontrado
         if(tecla_hex==tecla_hex_ant)portc=tecla_hex;
         return 1;//tom recebido 2 vezes
       }
     }
   }
   return 0;//tom no encontrado
}

//=============================================================================
//   INICIALIZAO DE PORTS E REGISTRADORES
//=============================================================================
void main (){
  //-----------------------------------------------------
  // PIC 16F676  definies de portas e registradores

  OPTION_REG=0B10000000; //timer 0 presc/2
  WPU=0;//sem pullups
  ANSEL = 0;   // desliga
  TRISA = 0b111111;  // port a direo dos pinos (in/out)
  TRISC = 0b100000;  // port c direo dos pinos (in/out)
  VRCON  = 0B10100000; //liga vref com 0 volts
  CMCON =  0B00000100;  // liga comparador interno - VREF=0 /NORMAL
  PORTC =   0b00000000;  // desliga sada
  zera_array();//zera acumuladores de senos/cossenos
  TMR0=0;
  INTCON= 0B10100000; //liga interrupo do timer 0 /
//=============================================================================
//   LOOP PRINCIPAL DE REPETIO DO PROGRAMA
//=============================================================================
  for(;;){
#define DELAY_STD_LIGADO   5

     while(T0IE);//aguarda finalizao de amostragens

     if(decodifica_dtmf()){;//procura por tom dtmf
       cnt_desl=0;//encontrou, e j acionou as sadas q1...q4
       std=1;//ativa a sada STD para indicar que recebeu tom
     }

     zera_array();//zera array de senos e cossenos
     TMR0=0;//zera timer 0
     cnt_amostragem=0;//reinicia contador de amostragem
     INTCON= 0B10100000; //liga interrupo do timer 0  e inicia nova amostragem
     if(++cnt_desl>DELAY_STD_LIGADO)std=0;//desliga sada STD se alcanou o tempo

  }//for(;;)
}//void main()
//-----------------------------------------------------------------------------


